# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Heritage SGRT (experimental) classes.
"""

import os

from geopathfinder.sgrt_naming import sgrt_path



class SgrtFolderName_old_interface():

    """
    Heritage-class from SGRT. This class handles sgrt data (folders)
    structure, using the geopathfinder package.
    """

    def __init__(self, product_id, wflow_id, dir_root,
                 dir_work=None, **kwargs):
        """
        Initializes sgrt data structure.

        Parameters
        ----------
        product_id : str
            Input satellite product ID. The full satalite product id has
            13 characters (e.g. S1AIWGRDH1HHD) at least 3 characters are
            required to identify the satellite (top level in folder structure)
        wflow_id : str
            A 5 digits string starting with a letter (A, B, C, or N)
            e.g. A0110 standing for wflow_name=A01, wflow_version=10
        dir_root : str
            The top level directory sgrt output folder structure will be
            created in that directory
        dir_work : str, optional
            Working directory (default: None).
        """

        self.all_levels = sgrt_path(
            dir_root=dir_root, product_id=product_id, wflow_id=wflow_id,
            **kwargs)

        self.level_1 = self.all_levels.build_levels(level='product')
        self.level_2 = self.all_levels.build_levels(level='group')

        self.dir_root = dir_root
        self.dir_anc = os.path.join(self.level_1, "ancillary")
        self.dir_doc = os.path.join(self.level_1, "documentation")
        self.dir_log = os.path.join(self.level_2, "logfiles")
        if dir_work is not None:
            self.dir_work = os.path.join(dir_work, ".SGRT")
        else:
            self.dir_work = os.path.join(dir_root, ".SGRT")

        self.product_id = product_id.upper()
        self.wflow_id = wflow_id.upper()

        self.build_subdirs(**kwargs)

    def build_subdirs(self, ptop_name=None, grid='EQUI7', ftile=None,
                      sgrt_var_name=None, makedir=False):
        """
        This function builds/updates directories according to the given
        parameters as keywords

        ptop_name : str, optional
            Product top directory (folder) name, in which the wflow
            outputs are stored The ptop_name name are actually coming from
            the workflow configuration files (e.g. ['geocoded', 'resampled']
            from A01_config.xml) NOTE: in the new version onyl ONE ptop_name
            will will be stored. In case of a list of ptop directrories,
            the code should be run in a loop
        grid : str, optional
            Grid reference for the data (default: 'EQUI7').
        ftile : str
            Full tile identifier e.g. 'EU010M_E052N016T1'
        sgrt_var_name : str
            The sgrt variable name is used to create the product
            sub directory name, in which the final product is stored
            NOTE: the ftile keyword should be already provided in order to
            create the product sub directory e.g. SIG0, PLIA, TMENPLIA, ..
        makedir (optional, default is false): flag
            a Flag for making sgrt directory
        NOTE: grid, ftile, and psub_dir (sgrt_var_name) directories will be
             created only if their parent directories are provided.
        NOTE: grid and ftile directories will be created if pir_sub
             sgrt_var_name) is also provided.
        """
        # check optional keywords to continue the directoy path to
        # level_3: product top directory
        # level_4: grid and tile name
        # level_5: product sub directory

        if ptop_name is None:
            raise ValueError(
                "You're using an outdated SGRT folder naming convention!")

        if ftile not in [None, '']:
            if len(ftile) != 17:
                raise ValueError(
                    "According to SGRT Naming Convention the full tile "
                    "name should have 17 characters!")

        if sgrt_var_name not in [None, '']:
            if len(sgrt_var_name) > 9:
                raise ValueError("According to SGRT Naming Convention the"
                                 "sgrt varible name should be not more"
                                 "than 9 characters!")

        levels = sgrt_tree(dir_root=self.level_2, ptop_name=ptop_name,
                           grid='EQUI7', ftile=ftile, wflow_id=self.wflow_id,
                           sgrt_var_name=sgrt_var_name, section='bottom')

        self.level_3 = levels.build_levels(level='wflow')
        self.level_4 = levels.build_levels(level='tile')
        self.level_5 = levels.build_levels(level='var')

        self.ptop_name = ptop_name
        self.dir_ptop = self.level_3
        self.dir_ftile = self.level_4
        self.dir_psub = self.level_5
        self.dir_qlook = levels.build_levels(level='qlook')

        if makedir:
            self.makedirs()

    def makedirs(self):
        """
        Make the directories that start with "dir"
        """
        for k, v in zip(self.__dict__.keys(), self.__dict__.values()):
            if k[0:3] == 'dir':
                makedirs(v)
            else:
                continue


class SGRTFolderName():

    """
    Heritage-class from SGRT. This class handles sgrt data (folders)
    structure, using the geopathfinder package.
    """

    def __init__(self, product_id, wflow_id, dir_root, dir_work=None,
                 **kwargs):
        """
        Initializes sgrt data structure.

        Parameters
        ----------
        product_id : str
            Input satellite product ID. The full satalite product id has
            13 characters (e.g. S1AIWGRDH1HHD) at least 3 characters are
            required to identify the satellite (top level in
            folder structure)
        wflow_id : str
            Work flow ID. A 5 digits string starting with a
            letter (A, B, C, or N) e.g. A0110 standing for wflow_name=A01,
            wflow_version=10.
        dir_root :
            The top level directory. Sgrt output folder structure will be
            created in that directory.
        dir_work : str, optional
            Working directory.
        """
        # function for the sensor and prodcut folders
        if product_id[0:2].lower() == 's1':
            sensor = "Sentinel-1_CSAR"
            product = product_id[3:9].upper()
        elif product_id[0:3].lower() == 'asa':
            sensor = "Envisat_ASAR"
            product = product_id[3:5].upper()
        elif product_id[0:3].lower() == 'asc':
            sensor = 'METOP_ASCAT'
            product = product_id[3:9].upper()
        elif product_id[0:3].lower() == 'sss':
            sensor = 'SCATSAR'
            product = product_id[3:8].upper()
        else:
            raise ValueError('product_id is unknown!')

        # function for the wflow_folder_names
        wflow_folder_names = {
            "A": "preprocessed", "B": "parameters", "C": "products",
            "N": "nrt"}

        wflow_group = wflow_folder_names[wflow_id[0].upper()]

        # function for datasets or logfiles
        datalog = 'datasets'

        # function for the ptop_name
        top_folder = ptop_name

        # function for wflow_id
        wflow = wflow_id

        # function for grid_folder, tile_folder
        grid_folder, tile_folder = "_".join(
            [grid.upper(), ftile.split('_')[0]]), ftile.split('_')[1]

        # function for the sgrt_var_name
        var_folder = sgrt_var_name.lower().rstrip('-')

        # function for qlooks
        qlook = 'qlooks'



def test_folders():
    root_path = r'R:\Projects_work\SAR_NRT_Code_Sprint\Testdata'

    ftf = sgrt_path(
        root_path, sensor='Sentinel-1_CSAR', mode='IWGRDH', group='products', datalog='datasets',
        product='ssm', wflow='C1003', grid='EQUI7_EU500M', tile='E048N012T6', var='ssm',
        make_dir=True)

    print(ftf['grid'])
    print(ftf.search_files('var', pattern=".*SSM.*VV.*", full_paths=True))
    files = ftf.search_files('var', pattern=".*SSM")

    df = ftf.search_files_ts('var', pattern=".*SSM", starttime='20161218_000000', endtime='20161224_000000',
                             full_paths=True)

    print(ftf.search_files('var', pattern='M*', full_paths=True))



if __name__ == '__main__':
    test_folders()



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


class FilenameObj(object):
    def __init__(self, attributes):
        for key, value in attributes.items():
            setattr(self, key, value)


class SmartFilename(object):

    """
    SmartFilename class handles file names with pre-defined field names
    and field length.
    """

    def __init__(self, fields, fields_def, ext=None, pad='-', delimiter='_'):
        """
        Define name of fields, length, pad and delimiter symbol.

        Parameters
        ----------
        fields : dict
            Name of fields (keys) and (values).
        field_def : OrderedDict
            Name of fields (keys) in right order and length (values).
        ext : str, optional
            File name extension (default: None).
        pad : str, optional
            Padding symbol (default: '-').
        delimiter : str, optional
            Delimiter (default: '_')
        """
        self.fields = fields
        self.fields_def = fields_def
        self.ext = ext
        self.delimiter = delimiter
        self.pad = pad
        self._check_def()
        self.obj = FilenameObj(self.fields)
        self.full_string = self._build_fn()


    def _check_def(self):
        """
        Check if fields names and length comply with definition.

        """
        for key, value in self.fields.items():
            if key in self.fields_def:
                if len(value) > self.fields_def[key]['len']:
                    raise ValueError("Length does not comply with "
                                     "definition: {:} > {:}".format(
                                         len(value), self.fields_def[key]['len']))
                if 'delim' not in self.fields_def[key].keys():
                    self.fields_def[key]['delim'] = True
            else:
                raise KeyError("Field name undefined: {:}".format(key))


    def _build_fn(self):
        """
        Build file name based on fields, padding and length.

        Returns
        -------
        filename : str
            Filled file name.

        """
        filename = ''
        for name, keys in self.fields_def.items():

            length = keys['len']
            delimiter = self.delimiter if keys['delim'] else ''

            if name in self.fields:
                if filename == '':
                    filename = self.fields[name].ljust(length, self.pad)
                else:
                    filename += delimiter + \
                        self.fields[name].ljust(length, self.pad)
            else:
                if filename == '':
                    filename = self.pad * length
                else:
                    filename += delimiter + self.pad * length

        if self.ext is not None:
            filename += self.ext

        return filename


    def get_field(self, key):
        '''
        Returns the string of the field with given key.

        Parameters
        ----------
        key : str
            name of the field

        Returns
        -------
        str
            part of the filename associated with given key

        '''
        field = self.fields[key]
        field_obj = getattr(self.obj, key)
        if (field_obj != field) and (len(field_obj) <= self.fields_def[key]['len']):
            field = field_obj
            self[key] = field

        return field.replace(self.pad, '')


    def __getitem__(self, key):

        return self.get_field(key)


    def __setitem__(self, key, value):

        if key in self.fields_def:
            if len(value) > self.fields_def[key]['len']:
                raise ValueError("Length does not comply with "
                                 "definition: {:} > {:}".format(
                                     len(value), self.fields_def[key]['len']))
            else:
                self.fields[key] = value
                setattr(self.obj, key, value.replace(self.pad, ''))
        else:
            raise KeyError("Field name undefined: {:}".format(key))


    def __repr__(self):

        return self._build_fn()



# Copyright (c) 2018, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
SGRT folder and file name definition.

"""

import os

from datetime import datetime
from collections import OrderedDict

from geopathfinder.folder_naming import SmartPath
from geopathfinder.file_naming import SmartFilename
from geopathfinder.folder_naming import build_smarttree
from geopathfinder.folder_naming import create_smartpath


# Please add here new sensors if they follow the SGRT naming convention.
allowed_sensor_dirs = ['Sentinel-1_CSAR',
                       'SCATSAR',
                       'METOP_ASCAT',
                       'Envisat_ASAR']


class SgrtFilename(SmartFilename):

    """
    SGRT file name definition using SmartFilename class.
    """

    def __init__(self, fields):

        self.date_format = "%Y%m%d"
        self.time_format = "%H%M%S"
        self.fields = fields.copy()

        if 'dtime_2' not in self.fields.keys():
            self.single_date = True
            apply_dtime_2 = False
        else:
            self.single_date = False
            apply_dtime_2 = True
            # if only daytime and no date is given
            if self.fields['dtime_2'].year < 1950:
                self.single_date = True

        fields_def = OrderedDict([
                     ('pflag', {'len': 1, 'delim': False}),
                     ('dtime_1', {'len': 8, 'delim': False}),
                     ('dtime_2', {'len': 8, 'delim': True}),
                     ('var_name', {'len': 9, 'delim': True}),
                     ('mission_id', {'len': 2, 'delim': True}),
                     ('spacecraft_id', {'len': 1, 'delim': False}),
                     ('mode_id', {'len': 2, 'delim': False}),
                     ('product_type', {'len': 3, 'delim': False}),
                     ('res_class', {'len': 1, 'delim': False}),
                     ('level', {'len': 1, 'delim': False}),
                     ('pol', {'len': 2, 'delim': False}),
                     ('orbit_direction', {'len': 1, 'delim': False}),
                     ('relative_orbit', {'len': 3, 'delim': True}),
                     ('workflow_id', {'len': 5, 'delim': True}),
                     ('grid_name', {'len': 6, 'delim': True}),
                     ('tile_name', {'len': 10, 'delim': True})
                    ])

        if 'dtime_1' in self.fields.keys():
            if self.single_date:
                date = self.fields['dtime_1'].strftime(self.date_format)
                if apply_dtime_2:
                    time = self.fields['dtime_2'].strftime(self.time_format)
                else:
                    time = self.fields['dtime_1'].strftime(self.time_format)
                self.fields['dtime_1'] = date
                self.fields['dtime_2'] = time
            else:
                self.fields['dtime_1'] = self.fields['dtime_1'].strftime(self.date_format)
                self.fields['dtime_2'] = self.fields['dtime_2'].strftime(self.date_format)

        super(SgrtFilename, self).__init__(self.fields, fields_def, pad='-', ext='.tif')


    def __getitem__(self, key):
        """
        Get field content.

        Parameters
        ----------
        key : str
            Field name.

        Returns
        -------
        item : str
            Item value.
        """
        item = super(SgrtFilename, self).__getitem__(key)

        if key == 'dtime_1':
            item = datetime.strptime(item, self.date_format)
        if key == 'dtime_2':
            if self.single_date:
                item = datetime.time(datetime.strptime(item, self.time_format))
            else:
                item = datetime.strptime(item, self.date_format)

        return item


    def __setitem__(self, key, value):
        """
        Set field content.

        Parameters
        ----------
        key : str
            Field name.
        value : str or datetime
            Field value.

        """
        if key == 'dtime_1' and isinstance(value, datetime):
            value = value.strftime(self.date_format)
        if key == 'dtime_2' and isinstance(value, datetime):
            if self.single_date:
                value = value.strftime(self.time_format)
            else:
                value = value.strftime(self.date_format)

        super(SgrtFilename, self).__setitem__(key, value)


def create_sgrt_filename(filename_string):
    """
    Creates a SgrtFilename() object from a given string filename

    Parameters
    ----------
    filename_string : str
        filename following the SGRT filename convention.
        e.g. 'M20170725_165004--_SIG0-----_S1BIWGRDH1VVA_146_A0104_EU500M_E048N012T6.tif'

    Returns
    -------
    SgrtFilename

    """

    helper = SgrtFilename({})
    filename_string = filename_string.replace(helper.ext, '')
    parts = filename_string.split(helper.delimiter)

    if not [len(x) for x in parts] == [9, 8, 9, 13, 3, 5, 6, 10]:
        raise ValueError('Given filename_string "{}" does not comply with '
                         'SGRT naming convention!')

    dtime_2_format = "%Y%m%d"
    if parts[1].endswith('--'):
        dtime_2_format = "%H%M%S--"

    fields = {
              'pflag': parts[0][0],
              'dtime_1': datetime.strptime(parts[0][1:], "%Y%m%d"),
              'dtime_2': datetime.strptime(parts[1], dtime_2_format),
              'var_name': parts[2],
              'mission_id': parts[3][0:2],
              'spacecraft_id': parts[3][2:3],
              'mode_id': parts[3][3:5],
              'product_type': parts[3][5:8],
              'res_class': parts[3][8:9],
              'level': parts[3][9:10],
              'pol': parts[3][10:12],
              'orbit_direction': parts[3][12:13],
              'relative_orbit': parts[4],
              'workflow_id': parts[5],
              'grid_name': parts[6],
              'tile_name': parts[7]
             }

    return SgrtFilename(fields)


def sgrt_path(root, mode=None, group=None, datalog=None,
              product=None, wflow=None, grid=None, tile=None, var=None,
              qlook=True, make_dir=False):
    '''
    Realisation of the full SGRT folder naming convention, yielding a single
    SmartPath.

    Parameters
    ----------
    root : str
        root directory of the path. must contain satellite sensor at toplevel.
        e.g. "R:\Datapool_processed\Sentinel-1_CSAR"
    mode : str
        e.g "IWGRDH"
    group : str, optional
        "preprocessed" or "parameters" or "products"
    datalog : str, optional
        must be "datasets" or "logfiles"
    product : str
        e.g. "ssm"
    wflow : str
        e.g. "C1003"
    grid : str
        e.g. "EQUI7_EU500M"
    tile : str
        e.g. "E048N012T6"
    var : str
        e.g. "ssm"
    qlook : bool
        if the quicklook subdir should be integrated
    make_dir : bool
        if the directory should be created on the filesystem

    Returns
    -------
    SmartPath
        Object for the path
    '''

    # check the sensor folder name
    if root.split(os.sep)[-1] not in allowed_sensor_dirs:
        raise ValueError('Wrong input for "root"!')

    # define the datalog folder name
    if datalog is None:
        if isinstance(wflow, str):
            datalog = 'datasets'
    elif datalog == 'logfiles':
        product = None
        wflow = None
        grid = None
        tile = None
        var = None
        qlook = False
    elif datalog == 'datasets':
        pass
    else:
        raise ValueError('Wrong input for "datalog" level!')


    # define the group folder name
    if group is None:
        if wflow.startswith('A'):
            group = 'preprocessed'
        elif wflow.startswith('B'):
            group = 'parameters'
        elif wflow.startswith('C'):
            group = 'products'
        else:
            raise ValueError('Wrong input for "wflow" level!')


    # defining the folder levels
    levels = [mode, group,
              datalog, product, wflow, grid,
              tile,  var, 'qlooks']

    # defining the hierarchy
    hierarchy = ['mode', 'group',
                 'datalog', 'product', 'wflow', 'grid',
                 'tile', 'var', 'qlook']

    if qlook is False:
        levels.remove('qlooks')
        hierarchy.remove('qlook')

    return create_smartpath(root, hierarchy=hierarchy, levels=levels,
                     make_dir=make_dir)


def sgrt_tree(root, target_level=None, register_file_pattern=None):

    '''
    Realisation of the full SGRT folder naming convention, yielding a
    SmartTree(), reflecting all subfolders as SmartPath()

    Parameters
    ----------
    root : str
        top level directory of the SGRT dataset, which is the sensor name in
        the SGRT naming convention.
        E.g.: "R:\Datapool_processed\Sentinel-1_CSAR"
    target_level : str, optional
        Can speed up things: Level name of target tree-depth.
        The SmartTree is only built from directories reaching this level,
        and only built down to this level. If not set, all directories are
        built down to deepest depth.
    register_file_pattern : str tuple, optional
        strings defining search pattern for file search for file_register
        e.g. ('C1003', 'E048N012T6').
        No asterisk is needed ('*')!
        Sequence of strings in given tuple is crucial!
        Be careful: If the tree is large, this can take a while!

    Returns
    -------
    SmartTree
        Object for the SGRT tree.
    '''

    # defining the hierarchy
    hierarchy = ['mode', 'group','datalog',
                 'product', 'wflow', 'grid',
                 'tile', 'var', 'qlook']

    # Check for allowed directory topnames for "root".
    if root.split(os.sep)[-1] in allowed_sensor_dirs:
        sgrt_tree = build_smarttree(root, hierarchy,
                                    target_level=target_level,
                                    register_file_pattern=register_file_pattern)
    else:
        raise ValueError('Root-directory "{}" does is '
                         'not a valid SGRT folder!'.format(root))

    return sgrt_tree



if __name__ == '__main__':
    pass
